/*******************************************************************************
 server.js: main server file

 $Author: dmc $
 $Revision: 2836 $
 $Date: 2020-10-29 15:55:01 -0300 (Thu, 29 Oct 2020) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/TSOLO-Carambola/server2.js $
 $Id: server2.js 2836 2020-10-29 18:55:01Z dmc $
*******************************************************************************/

var log4js = require('log4js');
log4js.setGlobalLogLevel('ERROR');
log4js.setGlobalLogLevel('INFO');
log4js.setGlobalLogLevel('DEBUG');
//log4js.setGlobalLogLevel('TRACE');

var _log = log4js.getLogger('server.js');

/*
firmware version number that will be fetched by server:
C2FW=06XXX
04: from 10-Jan-2020
05: from 15-Aug-2020: add autodetect of "ACK" from server - if so,
   we can then reset socket in the event of missing ACKs or otherwise-idle socket
06: 30-Sep-2020: no change in actual server.js firmware from -05, but add :280/cgi-bin commands
*/


// includes
var fs = require("fs"); // file system
var exec = require("child_process").exec;
eval(fs.readFileSync("./smt.common.lib/server_devices.js")+""); // defines get_carambola2_sta_connected()


    // Stuff for serial port ___________________________________________________________________
    global.SerialPort = require("serialport").SerialPort
    serialPort = new SerialPort("/dev/ttyATH0", {
        baudrate: 115200
    }, false); // this is the openImmediately flag [default is true]
    
    serialPort.open(function (error) {

        if ( error ) {
            _log.error('failed to open: '+error);
        } else {
            _log.info('serial port opened');
            serialPort.on('data', onSerialPortData);
            serialPort.on('readable', function () {
            	console.log("serialport readable");
            });
            serialPort.on('error', function () {
            	console.log("serialport error");
            });
            serialPort.on('close', function () {
            	console.log("serialport close");
            });
        }
    });

    var onSerialPortData = function(indata) {
        _log.trace(indata);
    	// LESSON #1 - don't do _log.debug() on a string that includes embedded nulls!!  Or it crashes, silently!!
        //_log.debug('receive serialport data...',indata.toString());
        logserialportactivity('serialport bytes received=',indata.length,printable_chars(indata.toString('ascii'),10));
        received_data = Buffer.concat([received_data,indata]);
//        var re_match = process_received_data(received_data.toString('ascii'),received_data);
//        if (re_match) {

        var re_match;
        while (re_match = process_received_data(received_data.toString('ascii'),received_data)) {
			if ('number' == typeof re_match) {
				//console.log(" trim nbytes=",re_match);
	        	received_data = received_data.slice(re_match);
			} else { // return value of a regexp.match()
/*
> 'xxxAT+CIPSTART="TCP","192.168.0.47",5000,1\r\n'.match('AT\\+CIPSTART="TCP","(\\d+\\.\\d+\\.\\d+\\.\\d+)')
[ 'AT+CIPSTART="TCP","192.168.0.47',
  '192.168.0.47',
  index: 3,
  input: 'xxxAT+CIPSTART="TCP","192.168.0.47",5000,1\r\n' ]
> 

*/

				//console.log(" trim nbytes=",re_match[0].length,"chars=",re_match[0],"[1]=",re_match[1]);
	        	received_data = received_data.slice(re_match.index+re_match[0].length); // [0] or [1]??
        	}
        }
    };

var last_logserialportactivity = new Date;
var interval_logserialportactivity = 5000;

function logserialportactivity(ay) {
	var now = new Date;
	//console.log("logserialportactivity",now - last_logserialportactivity, interval_logserialportactivity); return;
	if (now - last_logserialportactivity > interval_logserialportactivity) {
		var aa = Array.prototype.slice.call(arguments);
		aa.unshift('**serialportactivity**'); // prepend
		console.log.apply(console,aa); // 'console' needed here,
		last_logserialportactivity = now;
	}
}

function serialwrite(str) {
//	console.log("serialwrite",str);
	logserialportactivity("serialwrite bytes=",str.length,str);
	serialPort.write(str);
}

function serialwriteln(str) {
	//console.log("serialwriteln",str);
	logserialportactivity("serialwriteln bytes=",str.length,str);
	//console.log("\n\n");
	serialPort.write(str + "\r\n");
}

var responses = [];

function response_simple(str_match, str_reply, state_match) {
	return {
		re : "(?:^|\n)(" + str_match + ")(.*)",
		state : state_match, // undefined --> match all states
		// all states
		fixed_response : str_reply
	};
}

function response_fn(str_match, fn_reply, state_match) {
	return {
		// old and possibly buggy, don't think final (.*) is correct/needed:
//		re : "(?:^|\n)(" + str_match + ")(.*)", // non-capturing parentheses (?:
//		re : "(?:^|\n)(" + str_match + ")", // non-capturing parentheses (?:
// question: will this work?
		re : "(" + str_match + ")",
		state : state_match, // undefined --> match all states
		fn_response : fn_reply
	};
}



// https://www.espressif.com/sites/default/files/documentation/4a-esp8266_at_instruction_set_en.pdf

//responses.push(response_simple("AT\r\n","CARAMBOLA2\r\nOK\r\n")); // for MCUFW=105+
responses.push(response_simple("AT\r\n","BOTH\r\nOK\r\n")); // for MCUFW=105+, "do both"
//responses.push(response_simple("AT\r\n","OK\r\n")); // old MCU-fw 104-
responses.push(response_simple("AT\\+RFPOWER=\\d+\r\n","OK\r\n"));
responses.push(response_simple("AT\\+SLEEP=\\d+\r\n","OK\r\n"));
responses.push(response_simple("AT\\+CWAUTOCONN=\\d+\r\n","OK\r\n"));
responses.push(response_simple("AT\\+CWMODE=\\d+\r\n","OK\r\n"));
responses.push(response_simple("AT\\+CIPMODE=\\d+\r\n","OK\r\n"));
responses.push(response_simple("AT\\+CIPMUX=\\d+\r\n","OK\r\n"));

/*

*/


// JSON?
responses.push(response_fn("^{.*}\r\n", function (re_match) {
	var qq;
	try {
		qq = JSON.parse(re_match[1]); // TODO: span multiple lines
	} catch(e) {
		console.log("JSON ERROR=",e,"JSON=",re_match);
		return false;
	}
	//console.log("JSON match, parse=",qq);
	console.log("JSON OK:,",re_match[1]);
	return true;
}));

var cipstatus = 5;

/*
AT+CIPSTATUS
*/

responses.push(response_fn("AT\\+CIPSTATUS\r\n", function (re_match) {
	//console.log("Yuppers",re_match[0]);
	/*
	AT+CIPSTATUS
	Response
	STATUS:<stat>

	+CIPSTATUS:<link	ID>,<type>,<remote	IP>,<remote	port>,<local	port>,<tetype>
	Parameters
	• <stat>: status of the ESP8266 Station interface.
	‣ 2: The ESP8266 Station is connected to an AP and its IP is obtained.
	‣ 3: The ESP8266 Station has created a TCP or UDP transmission.
	‣ 4: The TCP or UDP transmission of ESP8266 Station is disconnected.
	‣ 5: The ESP8266 Station does NOT connect to an AP. 
	*/
	
	if (!ssid) {
		cipstatus = 5;
		serialwriteln("STATUS:"+cipstatus);
		return;
	}
	
	get_carambola2_sta_connected(function(is_connected) {
		if (is_connected) {
			if (cipstatus == 5) {
				cipstatus = 2;
			}
		} else {
			cipstatus = 5;
		}
		console.log("get_carambola2_sta_connected=",is_connected,"cipstatus",cipstatus);
		serialwriteln("STATUS:"+cipstatus);
	});
}));




var net = require('net');

var socket_to_server = new net.Socket();
var socket_to_server_connected = false;
var socket_to_server_trying_to_connect = false;
var ack_received = false;// to support FSOLO(007), we need to LEARN whether or not ACKs are sent
// unfortunately, FSOLO(007) does not ACK, therefore the usual situation in no inbound traffic;
// but for S7/etc, we DO send ACKs

var socket_in_timer = undefined, socket_in_timer_duration = 2*3000; // ORFer runs @ 3000ms

function arm_timeout_socket_in() {
	disarm_timeout_socket_in();
	// if we have received an ACK in the past, assume server is up for it:
	if (ack_received == true) { // console.log("***arm timer");
		socket_in_timer = setTimeout(fn_timeout_socket_in, socket_in_timer_duration);
	}
}

function disarm_timeout_socket_in() {
	if (!socket_in_timer) return;
	clearTimeout(socket_in_timer);
	socket_in_timer = undefined;
}

function fn_timeout_socket_in() {
	socket_to_server.destroy();
	_log.info('fn_timeout_socket_in Connection timeout --> destroy socket');
}

socket_to_server.on('data', function(indata) {
	//console.log("*** socket_to_server incoming data",indata.toString().trim());
//	console.log("*** socket_to_server incoming data",printable_chars(indata.toString('ascii'),10));
	_log.trace("server:incoming",indata.length,"bytes",printable_chars(indata.toString('ascii'),20));
	arm_timeout_socket_in(); // inactivity time for incoming data only
	// unfortunately, if SOLO sends shot data, it defers sending ORFs... so a steady stream of outgoing
	// traffic will result in ORF/ACK not happening.  **SO**, we need to arm_timeout_socket_in()

	if (indata.toString().trim() == 'ACK') {
//		console.log("$$$$ it's an ACK $$$$$ !!");
//		console.log("it's an ACK... ignoring!!");
		ack_received = true;
	}
	//console.log("************\n\n");
	serialwrite("+IPD," + indata.length + ":" + indata.toString());
});

socket_to_server.on('error', function(error,e2) { 
	_log.error('Connection error (' + error + ')');
	serialwriteln("");
	serialwriteln("ERROR");
	serialwriteln("CLOSED");


	// no need to do anything else; after 'error' fires, we will then receive a 'close',
	// and its handler will take care of restarting things
});

socket_to_server.on('close', function() {
	_log.info('Connection closed');
	socket_to_server_connected = false;
	socket_to_server_trying_to_connect = false;
	ack_received = false;
	if (cipstatus == 3) { // ‣ 3: The ESP8266 Station has created a TCP or UDP transmission.
		cipstatus = 4; // ‣ 4: The TCP or UDP transmission of ESP8266 Station is disconnected.
	}

});

socket_to_server.on('timeout', function() {
	socket_to_server.destroy();
	_log.info('Connection timeout --> destroy socket');
});

socket_to_server.on('drain', function() {
	_log.info('Connection drain');

});


var global_keepalive = false;
var global_nodelay = true;
var global_timeout = 17000;

socket_to_server.on('connect', function() {
	_log.info('Connection started','keepalive=',global_keepalive,'timeout=',global_timeout),'nodelay=',global_nodelay;
	socket_to_server_connected = true;
	socket_to_server_trying_to_connect = false;
	ack_received = false;
	cipstatus = 3; // ‣ 3: The ESP8266 Station has created a TCP or UDP transmission.
//	socket_to_server.setKeepAlive(global_keepalive ? true : false, 2000 ); // socket.setKeepAlive([enable][, initialdelay])
//console.log("force setKeepAlive = false"); socket_to_server.setKeepAlive(false); // socket.setKeepAlive([enable][, initialdelay])
	socket_to_server.setTimeout(global_timeout);
	socket_to_server.setNoDelay(global_nodelay);
	serialwriteln("CONNECT");
	serialwriteln("");
	serialwriteln("OK");
	ack_received = false; // clear it every time a new connection is madde
	arm_timeout_socket_in(); // inactivity timer
});

var http = require("http");





/*
AT+CIPSTART="TCP","192.168.0.47",5001,1

ERROR
CLOSED


AT+CIPSTART="TCP","192.168.0.47",5000,1
CONNECT

OK

*/
responses.push(response_fn("AT\\+CIPSTART=.*\r\n", function (re_match) {
	//console.log("Yuppers",re_match[0]);
	/* Etablish TCP Connection
	Set
	Command
		Single TCP connection (AT+CIPMUX=0):
			AT+CIPSTART=<type>,<remote	IP>,<remote	port>[,<TCP	keep	alive>]
	Response
		OK
		or
		ERROR


	*/
	if (!ssid) {
		serialwriteln("ERROR");
		return;
	}
	// can we extract the parms?
	var re2_match = re_match[0].match('AT\\+CIPSTART="TCP","(\\d+\\.\\d+\\.\\d+\\.\\d+)",(\\d+),(\\d+)');

	console.log("CIPSTART, socket_to_server_trying_to_connect=",socket_to_server_trying_to_connect);

	if (re2_match && !socket_to_server_trying_to_connect) {
		var ip_addr = re2_match[1];
		var port = re2_match[2];
		global_keepalive = re2_match[3];
		console.log("CIPSTART, IP=",ip_addr,port,global_keepalive);
		socket_to_server_trying_to_connect = true;
		// run the connection attempt
		socket_to_server.connect(port, ip_addr);
	} else {
		console.log("failed CIPSTART");
		serialwriteln("");
		serialwriteln("ERROR");
		serialwriteln("CLOSED");
	}
}));

/*
AT+CWJAP="SMTserverUplinkZZ","NULL"
...eventually produces  "FAIL\r\n"

AT+CWJAP="SMTserverUplink","NULL"

AT+CWJAP="SMTserverUplink","NULL"
CLOSED
WIFI DISCONNECT
WIFI CONNECTED
WIFI GOT IP

OK
*/

var ssid, password, previous_ssid, previous_password;

responses.push(response_fn("AT\\+CWJAP=.*\r\n", function (re_match) {
	// can we extract the parms?
	var re2_match = re_match[0].match('AT\\+CWJAP="(\\w+)","(\\w+)"');
	if (re2_match) {
		ssid = re2_match[1];
		var password = re2_match[2];
		if (password == 'NULL') password = '';
		var cmd = '';
		cmd += "uci set wireless.@wifi-iface[0].ssid='" + ssid + "'";
		cmd += "&& uci set wireless.@wifi-iface[0].key='" + password + "'";
		cmd += "&& uci set wireless.@wifi-iface[0].encryption='" + (password == "" ? "none" : "psk2") + "'";
		cmd += "&& uci commit && /etc/init.d/network reload";
		var nochange = false;
		if (ssid === previous_ssid && password === previous_password) {
			nochange = true;
		}
		
		get_carambola2_sta_connected(function(is_connected) {
//			console.log("new-CWJAP cmd=",cmd,"nochange=",nochange,"is_connected=",is_connected);
			// don't repeatedly change settings to same thing!  I suspect this causes delays/hangups associating.
			if (nochange && is_connected) {
				console.log("new-CWJAP cmd=",cmd,"nochange=",nochange,"is_connected=",is_connected);
				serialwriteln("OK"); // fake/quick success
			} else {
				previous_ssid = ssid;
				previous_password = password;
				exec( cmd, function(err,stdout,stderr) {
					console.log("new-CWJAP cmd=",cmd,"nochange=",nochange,"is_connected=",is_connected,"err=",!!err);
					if (err) {
						serialwriteln("ERROR");
					} else {
						serialwriteln("OK");
					}
				});
			}
		});

		//console.log("CWJAP ssid=",ssid,"password=",password);

	} else {
		serialwriteln("ERROR");
	}
}));


/*
AT+CWJAP?
+CWJAP:"SMTserverUplink","06:f0:21:45:7a:fe",6,-68

OK

--> SSID and MAC, channel number, sig strength
*/
responses.push(response_fn("AT\\+CWJAP\\?\r\n", function (re_match) {
	// We can't conveniently get channel number
	// we could get MAC but I don't think it matters
	//  --> iw dev wlan0  station dump | grep '^Station '  | head -1 | cut -d' ' -f2
	// all that really matters is signal strength

	get_carambola2_sta_sig_strength(function (err,sig) {
		if (err || !ssid || sig.trim().length == 0) {
			cipstatus = 5; // not connected
			serialwriteln("FAIL");
		} else {
			serialwriteln('+CWJAP:"' + ssid + '","11:22:33:44:55:66",6,' + sig);
			serialwriteln("");
			serialwriteln("OK");
		}
		console.log("carambola2_sta_sig_strength err=",err,"sig=",sig,"cipstatus=",cipstatus);
	});
}));


/*

AT+CIPSEND=36

OK
> 

*/

// cipsend_nbytes
cipsend_reading_bytes = false;

cipsend_timer = undefined;
cipsend_timer2 = undefined; // for detecting 

responses.push(response_fn("AT\\+CIPSEND=.*\r\n", function (re_match) {
	// can we extract the parms?
	var re2_match = re_match[0].match('AT\\+CIPSEND=(\\d+)\r\n');
	if (re2_match && cipstatus == 3) { // only accept SEND command if connection fully established
		cipsend_nbytes = +re2_match[1];
		// console.log("CIPSEND nbytes=",cipsend_nbytes);
		cipsend_timer = setTimeout(function() {
			_log.error("CIPSEND timeout");
		},1000);
		serialwrite("OK\r\n> \r\n");
		cipsend_reading_bytes = true;
	} else {
		serialwriteln("ERROR");
	}
}));

function cipsend_write(send_data/*Buffer*/) {
	clearTimeout(cipsend_timer);
	arm_timeout_socket_in(); // reset/re-arm
	//_log.info("server:write",send_data.length,"bytes",printable_chars(send_data.toString('ascii'),20));
	//
	_log.trace("server:write",send_data.length,"bytes",send_data.toString);
	socket_to_server.write(send_data);
	serialwriteln("Recv " + send_data.length + " bytes");
	serialwriteln("");
	serialwriteln("SEND OK");
}


//console.log("reponses=",responses);

var state = undefined;

function process_received_data(received_data/*as an ASCII string*/,received_data_buffer/* as Buffer */ ) {

	// if we are doing a CIPSEND.... have we picked up all the bytes yet?
	if (cipsend_reading_bytes && received_data_buffer.length >= cipsend_nbytes) {
		//b=console.log("**Need", cipsend_nbytes, "bytes, have",received_data.length);
//		var our_data = received_data.substr(0,cipsend_nbytes);
//		received_data = received_data.substr(cipsend_nbytes);
		var our_data = received_data_buffer.slice(0,cipsend_nbytes);
		cipsend_write(our_data);
		_log.debug("cipsend_reading_bytes=",our_data.length,"bytes");
		cipsend_reading_bytes = false;
//		return undefined;
		return cipsend_nbytes;
	}

	if (cipsend_reading_bytes) {
		_log.trace("Need", cipsend_nbytes, "bytes, have",received_data.length);
		return;
	}

	var re_match;
	for (var i in responses) {
		response = responses[i];
		if (state === response.state || response.state === undefined) {
			if (re_match = received_data.match(response.re)) {
				var fn_return_value = undefined;
				_log.info("matched",re_match[0].trim() );
				if (response.fixed_response) {
					serialwrite(response.fixed_response);
				} else if ('function' == typeof response.fn_response) {
					fn_return_value = response.fn_response(re_match);
				}
				if (fn_return_value != false) {
					//console.log("return re_match for", response.re);
					return re_match;
				} else {
					//console.log("*** NOT returning re_match for", response.re);
					// fallthrough and try other patterns
				}
			}

		}
	}

	if (received_data.match(/.*\r\n/)) {
		_log.error("no match for",printable_chars(received_data));
	}
	return undefined;

}

function printable_chars(input,max_chars/*optional*/) {
	var out = input.replace(/[^\x20-\x7E]+/g, ""); // only printable chars
var out = input.replace(/[^\r\n\x20-\x7E]+/g, ""); // DEBUG: \r, \n and printable chars
	if ('number' == typeof max_chars) out=out.substring(0,max_chars);
	out = '(printable):' + out;
	return out;
}


var received_data = Buffer("");
