/*******************************************************************************
 $Author: tobias $
 $Revision: 3007 $
 $Date: 2021-10-05 05:12:24 -0300 (Tue, 05 Oct 2021) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smt.common.lib/fa-server.js $
*******************************************************************************/

global.is_R5 = false;

// set the log level for the log4js library
var log4js = require( "log4js" );
var os = require( "os" );


var net_ping = undefined, ping_session = undefined;
try {
	net_ping = require ("net-ping");
	console.log("net-ping YES");
} catch(err) {
	console.log("net-ping NO");
	; // ignore error, just don't use net_ping
}

if (net_ping && process.getuid() == 0) { // only root can run ping
	ping_session = net_ping.createSession( { retries:0, timeout:1000 } );
}

var debug_pinger = true;
debug_pinger = false;

function Pinger() {
	this.name = "Pinger";
	this.time = undefined;
	this.pinger_itimer = undefined;
	this.reset_stats();
}

Pinger.prototype.reset_stats = function() {
	this.samples = [];
	this.replies_rcvd = [];
	this.replies_lost = [];
	this.replies_lost_count = 0; // in our sliding sample
	this.replies_max = 120; //prod'n - two minutes
	//this.replies_max = 25; // for debug
	this.replies_slow_connection = 15; // threshold at which connection is reported to be "SLOW"
	this.most_recent_reply = undefined;
	this.reset_time_limit = 1000 * 2 * this.replies_max;
	// for all pings since start
	this.total_pings_lost = 0;
	this.total_pings_rcvd = 0;
	this.total_pings_min = 0;
	this.total_pings_max = 0;
	this.total_pings_mean = 0;
	this.total_pings_sd = this.total_pings_sd2 = 0;
}

Pinger.prototype.check_reset = function() {
	var now = new Date();
	if (this.most_recent_reply == undefined || now < this.most_recent_reply || now - this.most_recent_reply  > this.reset_time_limit) {
		this.most_recent_reply = undefined;
		this.replies_rcvd = [];	this.replies_lost = []; this.replies_lost_count = 0;
		this.reset_at_time = Date();
	}
}

Pinger.prototype.get_health = function() {
	var stats = calc_mean_sd(this.samples);
	// 	return {mean:_avg, sd: _sd, min: _min, max: _max, stderr: _sd/Math.sqrt(samples.length),  meanrange:_mr };
	var rtn = {
		show: "", // text to show
		description:"",
		replies: this.replies_rcvd.length,
		replies_lost_count: this.replies_lost_count,
		since_lost: this.mr_packet_loss_time ? Math.round((Date.now()-this.mr_packet_loss_time)/1000) : NaN,
		since_ping_rcvd: this.mr_ping_rcvd_time ? Math.round((Date.now()-this.mr_ping_rcvd_time)/1000) : NaN,
		ping_mean_time: +(stats.mean).toFixed(1),
		ping_stdev: +(stats.sd).toFixed(2),
		ping_min: stats.min, ping_max: stats.max,

		total_pings_lost: this.total_pings_lost,
		total_pings_rcvd: this.total_pings_rcvd,
		total_pings_min: this.total_pings_min,
		total_pings_max: this.total_pings_max,
		total_pings_mean: +this.total_pings_mean.toFixed(2),
		total_pings_sd: +this.total_pings_sd.toFixed(2),
		reset_time_limit: this.reset_time_limit,
		reset_at_time: this.reset_at_time,
		my_ip: this.my_ip,
	};

	if (this.replies_rcvd.length < this.replies_max) {
		rtn.description = "gathering data..";
		rtn.show = "--";
		rtn.show = '(' + parseInt(100 * this.replies_rcvd.length / this.replies_max)+'%)';
	} else if (this.replies_lost_count != 0) {
		rtn.description = this.replies_lost_count + " lost packets";
		rtn.show = "LOSS";
	} else if (rtn.ping_mean_time > this.replies_slow_connection) {
		rtn.description = "slow, mean="+(rtn.ping_mean_time).toFixed(0);
		rtn.show = "SLOW";
	} else {
		rtn.show = "OK";
	}
	if (global.data.config.is_SMT) {
		rtn.show += '<br><font size="-0">';
		// percentage packet success
		rtn.show += (100*this.total_pings_rcvd/(this.total_pings_rcvd+this.total_pings_lost)).toFixed(1)+'% (' + 
			this.total_pings_lost + ' lost)';
		rtn.show += '<br>';
		rtn.show += this.total_pings_min + '/' + this.total_pings_max;
		rtn.show +=  ' ' + (rtn.total_pings_mean).toFixed(0) + '/<b>' + (rtn.total_pings_sd).toFixed(1) + '</b>';
		rtn.show += '</font>';
	}
	return rtn;
}

Pinger.prototype.start = function(target_ip,rate_ms) {
	this.stop();
	this.reset_stats();
	this.my_ip = target_ip.replace(/::ffff:/,''); // '::ffff:192.168.0.215' --> '192.168.0.215'
	rate_ms = rate_ms || 1000;
	var me = this;
	this.pinger_itimer = setInterval(function() {
		//console.log("dmc itimer me.my_ip=",me.my_ip);
		me.sendping.call(me,me.my_ip);
	
	},rate_ms,target_ip);
}

Pinger.prototype.stop = function() {
	if (this.pinger_itimer) clearInterval(this.pinger_itimer);
}


Pinger.prototype.gen_pingresponse_fn = function () {
	var me = this;
	return function(error, target, sent, rcvd) {
		var pingtime = undefined;
		/*
		exports.DestinationUnreachableError = DestinationUnreachableError;
		exports.PacketTooBigError = PacketTooBigError;
		exports.ParameterProblemError = ParameterProblemError;
		exports.RedirectReceivedError = RedirectReceivedError;
		exports.RequestTimedOutError = RequestTimedOutError;
		exports.SourceQuenchError = SourceQuenchError;
		exports.TimeExceededError = TimeExceededError;
		*/
		if (error instanceof net_ping.RequestTimedOutError
		|| error instanceof net_ping.DestinationUnreachableError
		) {
			//console.log("timeout", target, sent- rcvd,error);
		} else if (!error) {
			me.most_recent_reply = rcvd;
			pingtime = rcvd - sent;
			if (pingtime < 0 || pingtime > 60000) { // sanity checks
				pingtime = undefined;
			}
		} else { // other error
			if (debug_pinger) console.log("other error - ",error);
		}
		me.replies_rcvd.push(pingtime); // lost -> undefined
		if (me.replies_rcvd.length > me.replies_max) {
			var outping = me.replies_rcvd.shift();
		}
		me.samples = [];
		var i; for (i = 0; i < me.replies_rcvd.length; i++) {
			var thispingtime = me.replies_rcvd[i];
			if (undefined !== thispingtime) {
				me.samples.push(thispingtime);
			}
		}

		var this_packet_lost;
		if (pingtime === undefined) {
			this_packet_lost = 1;
			me.total_pings_lost += 1;
			me.mr_packet_loss_time = Date.now();
		} else { // ping received; update stats
			this_packet_lost = 0;
			if (me.total_pings_rcvd == 0) {
				me.total_pings_min = me.total_pings_max = pingtime;
			} else {
				if (pingtime < me.total_pings_min) me.total_pings_min = pingtime;
				if (pingtime > me.total_pings_max) me.total_pings_max = pingtime;
			}
			var nm1 = me.total_pings_rcvd;
			me.total_pings_rcvd += 1;
			var n = me.total_pings_rcvd; var xn = pingtime;
			var xbarnm1 = me.total_pings_mean;
			var xbarn = (xbarnm1*nm1 + xn) / n;
			var varnm1 = me.total_pings_sd2;
			var varn = 0; // sdn^2
			// https://math.stackexchange.com/questions/102978/incremental-computation-of-standard-deviation
			if (n > 1) {
				varn = (n-2)/(n-1)*varnm1 + Math.pow(xn-xbarnm1,2)/n;
			}
			me.total_pings_sd = Math.sqrt(varn);
			me.total_pings_sd2 = varn;
			me.total_pings_mean = xbarn;

			me.mr_ping_rcvd_time = Date.now();
		}
		me.replies_lost.push(this_packet_lost); // lost -> undefined
		if (me.replies_lost.length > me.replies_max) me.replies_lost.shift();
		var i; for (i = 0, me.replies_lost_count = 0; i < me.replies_lost.length; i++) {
			me.replies_lost_count += me.replies_lost[i];
		}
		
		if (debug_pinger) console.log("pingresponse()", error ? "error" : "ok"
			,"thisping=",error ? undefined : pingtime
			,"replies=",me.replies_rcvd.length,"/",me.replies_max
			,"lost=",me.replies_lost_count
		)
	}
}

Pinger.prototype.sendping = function(target_ip) {
	this.check_reset();
	ping_session && ping_session.pingHost(target_ip, this.gen_pingresponse_fn());
}

var logAppenders = [
    {
        "type": "logLevelFilter",
        "level": "WARN",
        "appender": {
                type: "console"
            }
    }



];

log4js.configure( {
    appenders: logAppenders
} );

// default global loglevel is set in the appender with no 'category'
// Don't use these:
//log4js.setGlobalLogLevel( "ERROR" );
//log4js.setGlobalLogLevel( "DEBUG" );
//log4js.setGlobalLogLevel( "TRACE" );

var net = require( "net" );

var log = log4js.getLogger( "fa-server.js" );

var shots = [],
    shotCounter = 0,
    lastServerPort = 11001;

log.info( "fa-server starting..." );

// Possible TODO:
// conn.on( "get-shot", function( data ) {
//     var index = parseInt( data.data.number, 10 );
//     if ( shots[ index ] !== undefined ) {
//         conn.send( data.socket, "get-shot", { result: true, shot: shots[ index ], number: index } );
//     } else {
//         conn.send( data.socket, "get-shot", { result: false, shot: undefined, number: index  } );
//     }
// } );


// stuff for the fa server (the fa connects to a server)
// at the moment the FA tries to connect to port 5000
var clients = [],
    faServerPort = 5000;

global.fa_clients = clients;

// https://gist.github.com/creationix/707146
net.createServer( function( socket ) {
    socket.name = socket.remoteAddress + ":" + socket.remotePort;
    var MessageHandler = require( "./lib/messagehandler" );

    var client = undefined;

    // search existing socket... reuse client if possible
    for ( var i = 0; i < clients.length; i++ ) {
        if ( clients[ i ].ip === socket.remoteAddress) {
            if (!client) client = clients[i]; //re-use the first one we find
            if (clients[i].active) {
                clients[ i ].active = false;
                logStatistic( clients[ i ], "LP" );
                logStatistic( clients[ i ], "D" );
                clients[ i ].socket.destroy(); // clean up any other incoming connections from same IP
                // leave pinger alone, so that it can span disconnect-connect events
            }
        }
    }

    fa_server_log("fa-server new connection; socket.setNoDelay(), clients.length=",clients.length);
    socket.setNoDelay(); // so socket.write() is delivered immediately

    // var client = {
    //     socket: socket,
    //     ip: socket.remoteAddress,
    //     port: socket.remotePort,
    //     last: Date.now(),
    //     start: Date.now(),
    //     heartbeats: 0,
    //     active: true,
    //     firstHeartbeat: undefined,
    //     lastHeartbeat: undefined,
    //     localIp: socket.address().address,
    //     /* fields which will be filled in by orf_updater
    //     remote_rf_level, temperature, serialno
    //     connected, ORF->true, ORF-TIMEOUT -> false
    //     */
    //     circBufSize: 46, // will be set by shot-fa
    //     fn_send_huhu: send_huhu,
    //     fn_request_shots: request_shots,
    //     fn_refetch_needed: refetch_needed,
    //     fn_do_refetch: do_refetch,
    //     get_shot_number_mod: get_shot_number_mod,
    // };

    if (!client) {
        client = {}; // new
		client.pinger = new Pinger;
        clients.push( client );
    } else {
    	// reused client uses the pinger already there
        fa_server_log("fa-server re-use client, client.most_recent_shot_no=",client.most_recent_shot_no);
    }
    client.socket = socket;
    client.ip = socket.remoteAddress;
    client.port = socket.remotePort;
    client.last = Date.now();
    client.start = Date.now();
    client.heartbeats = 0;
    client.active = true;
    client.firstHeartbeat = undefined;
    client.lastHeartbeat = undefined;
    client.localIp = socket.address().address;
    if (client.itimer) clearInterval(client.itimer);
    client.itimer = undefined;
    /* fields which will be filled in by orf_updater
    remote_rf_level, temperature, serialno
    connected, ORF->true, ORF-TIMEOUT -> false
    */
    // client.circBufSize = 48; // will be set by shot-fa
    client.circBufSize = undefined; // will be set by shot-fa
    client.fn_send_huhu = send_huhu;
    client.fn_request_shots = request_shots;
    client.fn_refetch_needed = refetch_needed;
    client.fn_do_refetch = do_refetch;
    client.get_shot_number_mod = get_shot_number_mod;
    client.pinger.start(client.ip);
    fa_server_log("fa-server client.ip/port=",client.ip,client.port/*," pinger=",client.pinger*/);


    // // try to find an existing device for the given ip...
	// for ( var i in global.data.devices ) {
    //     var d = global.data.devices[ i ];
    //     if ( d.type === "SOLO" && d.deviceIP === client.ip ) {
    //         client.device = d;
    //         break;
    //     }
    // }


    logStatistic( client, "C" );

    // handler for incoming shots
    var shotFA = new MessageHandler.ShotHandlerFA();
    shotFA.on("new-shot", function (shotModel) {
        client.circBufSize = shotModel.circBufSize;
        fa_server_log(client.circBufSize,'-',"fa-server new_shot", shotModel.shotNumber, "client=", client.serialno, shotModel.shotCountVector, client.next_shot_no);
        // bugfix1 - before first ORF is sent, client.serialno = client.device = undefined; therefore shots
        // handed to "measFromBeam" will be discarded
        if (!client.device) {
            fa_server_log("BUGFIX1, new-shot: most_recent=",client.most_recent_shot_no,'MR-1=',client.get_shot_number_mod(shotModel.shotCountVector-1),'!client.device=',!client.device,'client.device=',client.device);
            if (client.most_recent_shot_no == undefined) {
            	client.most_recent_shot_no = client.get_shot_number_mod(shotModel.shotCountVector-1);
            }
            //fa_server_log('shotModel=',shotModel);
            fa_server_log("BUGFIX1... ignoring for now; refetch will happen later");
            return;
        }
        if (client.fn_refetch_needed(shotModel.shotCountVector)) {
            fa_server_log("new-shot: refetch needed!");
            if (global.data && global.data.config && global.data.config.solo_test_no_refetch_new_shot) {
                fa_server_log("... but, ignoring");
                ; // ignore error, carry on
            } else {
                fa_server_log("... discard & refetch");
                client.fn_do_refetch(); // ask for shots to be re-sent, ignore this shot for now, we
                return; // will later request it to be resent and we'll deal with it then.
            }
        }

        var shot = {
            shotnum: shotModel.shotNumber, // 0..65535
            counters: [
                shotModel.counter0,
                shotModel.counter1,
                shotModel.counter2,
                shotModel.counter3,
                shotModel.counter4,
                shotModel.counter5,
                shotModel.counter6,
                shotModel.counter7
            ],
            raw_counters: shotModel.raw_counters,
            G2_temp_rdg: shotModel.mean_temp_uncal,
            out_set: 0,
            GPIO18: shotModel.GPIO18, // JRPL target position reading
            mcucount: shotModel.shotCountVector // 0..46(e.g.)
        };
        if (global.data.config.global_temperature !== undefined) {
            shot.G2_temp_rdg = global.data.config.global_temperature;
        }

        var logshot = {
          ts: Date.now(),
          source: "SOLO",
          shot: shot
        };

        log.trace( "LOGSHOT = ", logshot );
        if (global.data && global.data.config && global.data.config.solo_test_discard_shots > 0) {
			global.data.config.solo_test_discard_shots--;
			fa_server_log("fa-server solo_test_discard_shots #",shotModel.shotCountVector
			  ,'mr_shot=',client.most_recent_shot_no, 'next_shot_no=',client.next_shot_no
			  ,'discards remaining:',global.data.config.solo_test_discard_shots
			);
			; // ignore shot, do not process
        } else {
            // When we receive a shot, the previously-reported next_shot_no might be out of date,
            // so clear it; it will be correctly set by the next ORF.
            client.most_recent_shot_no = shotModel.shotCountVector;
            client.next_shot_no = undefined; // wait until the next ORF
            fa_server_log("DMC fa-server process shot");
            global.qndEventEmitter.emit( "measFromBeam", {
                beamMeasurement: logshot,
                srcDevice: client.device
            } );
        }
    } ); // fired from lib/connection/tcp/index.js:206

    // handler for incoming ORFs
    var orfFA = new MessageHandler.OrfHandlerFA();

    var my_orf_updater = (function (my_client) {
        return function (orf_msg) {
            var cmd = new Buffer("ACK\r\n");
            // allow GUI to turn off ACK...
            var no_ack = false;
            if (my_client && my_client.device && my_client.device.SOLO_reset) {
            	no_ack = true;
				var crud_msg = {devices: {}};
				crud_msg.devices[my_client.device.serialno] = {SOLO_reset: false};
				apply_crud(crud_msg);
            }
            if (no_ack) {
			    fa_server_log("fa-server NOT sending ACK"); // fw version 103 will respond by doing an ESP RESET
        	} else {
			    fa_server_log("fa-server send ACK cmd=", cmd, "ASCII=", cmd.toString().trim());
			    my_client.socket.write(cmd);
        	}

            log.trace("my_orf",orf_msg,my_client.ip);
            var exists = false;
            my_client.serialno = orf_msg.serialno;
            my_client.remote_rf_level = orf_msg.remote_rf_level;
            my_client.heartbeats++;
            my_client.next_shot_no = orf_msg.next_shot_no;
            my_client.connected = true; // ORF means we are connected
            my_client.mean_temp_uncal = orf_msg.mean_temp_uncal;
            var hrt_start = my_client.orf_hrtime; // previous value, if any
            var hrt_end = my_client.orf_hrtime = process.hrtime();
            log.debug('orf!',client==my_client, 'SN=',my_client.serialno, 'nxtShot=', my_client.next_shot_no
                , '(mr_shot=' + (client.most_recent_shot_no === undefined ? 'undefined' : client.most_recent_shot_no) + ')'
                , 'RF=',orf_msg.remote_rf_level, 'mean_temp_uncal=',orf_msg.mean_temp_uncal
            );


            if (hrt_start) {
                var hr_et = +( (hrt_end[0]-hrt_start[0])*1000+(hrt_end[1]-hrt_start[1])/1000000 ).toFixed(0);
                // maintain an exponential moving average orf-time
                // only change for "reasonable" and stable orf-times
                if (hr_et > 2000 && hr_et < 20000 && my_client.heartbeats > 3) {
                    var exp = 1 / 5, scaleup = 1.20; // exp-decay 1/5th, wait for 20% longer than typical
                    if (0 == orfFA.orf_timeout_value) {
                        orfFA.orf_timeout_value = hr_et * scaleup;
                    } else {
                        orfFA.orf_timeout_value = (1 - exp) * orfFA.orf_timeout_value + exp * hr_et * scaleup;
                    }
                }
                log.trace("ORF->ORF time = ",hr_et,"ms, orfFA.orf_timeout_value=",orfFA.orf_timeout_value,my_client.heartbeats);
            }
            
            // 12-Mar-2019: change to always andimmediately do this when ORF is received
            emitBroadcast(my_client); // will add my_client to global.data.devices if it isn't already there

            // try to find an existing device for the given ip...
            // should ALWAYS succeed
            log.trace('orf_updater, searching for device');
            for ( var i in global.data.devices ) {
                var d = global.data.devices[ i ];
                if ( d.type === "SOLO" && d.serialno === orf_msg.serialno ) {
                    my_client.device = d;
                    exists = true;
		            log.trace('orf_updater, FOUND device');
                    break;
                }
            }
            if (!exists) {
            	log.error('orf_updater SOLO=',orf_msg.serialno,'device not found');
            }

            if (client.fn_refetch_needed()) {
                fa_server_log("orf_updater: refetch needed!");
                if (global.data && global.data.config && global.data.config.solo_test_no_refetch_orf) {
                    fa_server_log("... but, ignoring"); // log it, don't refetch, carry on with ORF processing
                } else {
                    fa_server_log("... refetching");
                    client.fn_do_refetch(); // ask for shots to be re-sent, ignore this shot for now, it will come later
                    ; // carry on with ORF processing.  VERY IMPORTANT that we run the "search for device" code below.
                }
            }

        }
    })(client); // closure

    var my_orf_timeout_updater = (function (my_client) {
        return function (){
            fa_server_log("ORF-TIMEOUT ",my_client.serialno);
            my_client.connected = false; // ORF-TIMEOUT
            emitBroadcast(my_client); // immediately announce device is down
        }
    })(client); // closure

    orfFA.on( "ORF", my_orf_updater); // fired from messagehandler orf-fa.js
    orfFA.on( "ORF-TIMEOUT", my_orf_timeout_updater); // fired from messagehandler orf-fa.js

    var proxy = new MessageHandler.Proxy(); // hardcoded in there is the knowledge that 0xBB is our framing byte
    proxy.addHandler( 0xAA, shotFA ); // 0xaa is a new shot
    proxy.addHandler( 0xC0, shotFA ); // 0xc0 is a shot that we requested
    proxy.addHandler( 0xdc, orfFA ); // DMC - OK to use 0xdc as message type for ORF???


	if (1) { // instead
		// huhu will tell us fwversion and hwversion

		setTimeout( function() {
			fa_server_log("one-off fn_send_huhu()",client.serialno, client.remote_rf_level);
			client.fn_send_huhu();
		}, 15 * 1000); // OK

		client.itimer = setInterval( function() {
			fa_server_log("interval fn_send_huhu()",client.serialno, client.remote_rf_level, "client.hwversion",client.hwversion,"client.fwversion",client.fwversion);
			client.fn_send_huhu();
		}, 80 * 1000);
	}


    // Handle incoming messages from clients.
    socket.on( "data", function( data ) {

        client.last = Date.now();

        var str = data.toString( "ascii" ); // 'ascii' gives byte-for-byte conversion
        var buf = new Buffer( data );
        var numOrfBytes = orfFA.numberOfDataBytes; // likely 64
        var orfRegexp = orfFA.orfRegexp;
        var orfRegexp_huhu = orfFA.orfRegexp_huhu;

        log.trace( "RAW DATA IN " + socket.remoteAddress + ":" + socket.remotePort + ": " + buf.toString( "hex" ) );
        //log.trace( "RAW DATA str= ", str )
        if (!client.active) {
            return;
        }

        orfFA.arm_orf_timeout(); // on *ANY* communication, reset the ORF-timeout

        var re_match_orf = str.match(orfRegexp); // unsolicited ORFs match this pattern
        var re_match_huhu = str.match(orfRegexp_huhu); // ORFs from huhu match this different pattern
        
        if (re_match_huhu) {
        	client.hwversion = parseInt(re_match_huhu[1]);
        	client.fwversion = parseInt(re_match_huhu[2]);
            log.debug('data re_match_huhu',client.serialno, 'hw=', client.hwversion, 'fw=', client.fwversion);
            log.trace(re_match_huhu);
        } else if(re_match_orf) { // we have an ORF, and maybe some preceding and following data

//fa_server_log("orf --> fn_send_huhu()",client.serialno, client.remote_rf_level, "client.hwversion",client.hwversion,"client.fwversion",client.fwversion);
//client.fn_send_huhu();



            // console.log("DMC, orf-regexp", re_match,"numberOfDataBytes",orfFA.numberOfDataBytes);

//> 'sdsfdORF-123-4095,4095,123,91,-54\r\r\nblsdfkjdsf'.match(/ORF-([^-]+)-(\d+),(\d+),(\d+),(\d+),(-{0,1}\d+)\r*\n/)

/*
[ 'ORF-123-4095,4095,123,91,-54\r\r\n',
'123',
'4095',
'4095',
'123',
'91',
'-54',
index: 5,
input: 'sdsfdORF-123-4095,4095,123,91,-54\r\r\nblsdfkjdsf' ]
>

*/


            proxy.addRawData(buf.slice(0, re_match_orf.index)); // chars preceding ORF

            // orfbit -- add framing
            proxy.addRawData(Buffer([0xbb, 0xbb, 0xbb, 0xbb, 0xdc])); // DMC TEMP.. ok for 0xdc???
            proxy.addRawData(Buffer(re_match_orf[0]).slice(0,numOrfBytes)); // matched string; might be less than numOrfBytes (64?) bytes
            var match_len = re_match_orf[0].length;
            var padBuffer = new Buffer(numOrfBytes - match_len);
            padBuffer.fill(' ');
            proxy.addRawData(padBuffer); // now we've written exactly numOrfBytes(64?) bytes for the ORF

            proxy.addRawData( buf.slice( re_match_orf.index + match_len ) ); // write any following characters
            // orf-fa.js:orfFA() will handle this data, based on message type 0xdc

            // log.trace( "fixed-up ORF: " +  fixed_orf );
        } else { // no ORF, or corrupted enough that I am not willing to handle it!!
            proxy.addRawData(data);
        }

//         /// ORFs are not framed; so we must work around this here.
//         // ASSUMPTIONS:
//         // -- there is no useful (i.e. shot) data prior to the "ORF"
//         // -- the ORF ends with a newline \n, and there might be useful data after it
//         // we need to force serialno to exactly four digits, padded.

//         var orfset = str.indexOf( "ORF-" );
//         var orfstr = str.slice(orfset);
//         // var dash1 = orfstr.indexOf('-'); // offset of first dash, otherwise -1
//         // var dash2 = dash1 == -1 ? dash1 : orfstr.indexOf('-',dash1+1); // position of second dash
//         var endorf = orfstr.indexOf('\n'); // might not be there
//         if (endorf < -1) endorf = orfstr.length;
//         console.log("DMC orf-look endorf=",endorf,"orfset=",orfset,"orfFA.numberOfDataBytes=",orfFA.numberOfDataBytes);

//         log.trace("DMC - trace??")
// //        log.debug("DMC:!!!!!!!!!\n\n\n",str,orfset,dash1,dash2,endorf,"***********\n\n!!!!")
//         log.trace("DMC - trace---")

//         if ( orfset < 0 || /* dash1 < 0 || dash2 < 0 || */ endorf < 0) { // no ORF, or corrupted enough that I am not willing to handle it!!
//             proxy.addRawData( data );
//         } else { // we have an ORF, with BOTH dashes and the newline
//             // data before-ORF, with-ORF, and after-ORF
//             var firstbit = orfstr.slice(0,orfset); // data preceding our ORF-
//             var orfbit = orfstr.slice(orfset,endorf<0?undefined:endorf); // we will re-write this to exactly 34 bytes, and add framing
//             //var endbit = orfstr.slice(endorf+1);

//             // create a fixed-up ORF
//             // var serialno = orfstr.slice(dash1+1,dash2);
//             // var fixed_orf = orfstr.slice(0,dash1+1) + ('0000'+serialno).slice(-4);
//             var fixed_orf = orfstr.slice(0,endorf);
//             fixed_orf += new Array(65).join(' '); // pad
//             fixed_orf = fixed_orf.slice(0,64); // make exactly 64 chars
// console.log("DMC: orf in = ", orfstr, "orfbit=",orfbit,". serialno=",serialno,".");
// console.log("DMC: fixed_orf1 = ", fixed_orf);



//             if (endorf < 0) fixed_orf += orfstr.slice(dash2);
//             else  fixed_orf += orfstr.slice(dash2,endorf);
// //console.log("DMC: fixed_orf = ", fixed_orf);

//             proxy.addRawData( buf.slice( 0, orfset ) ); // firstbit

//             // orfbit
//             proxy.addRawData( Buffer([0xbb, 0xbb, 0xbb, 0xbb, 0xdc] ) ); // DMC TEMP.. ok for 0xdc???
//             //proxy.addRawData( buf.slice( orfset ) );
//             proxy.addRawData( Buffer( fixed_orf ) );

//             // The endbit won't exist if the ORF was not terminated with \n
//             if (endorf >= 0) proxy.addRawData( buf.slice( endorf+1 ) );

//             log.trace( "data received ascii=" + str );
//             log.trace( "fixed-up ORF: " +  fixed_orf );
// //            log.trace( "'heartbeat' received" );
//         }

        if ( client.firstHeartbeat === undefined ) {
            client.firstHeartbeat = Date.now();
            client.lastHeartbeat = Date.now();
            logStatistic( client, "FP" );
        } else {
            client.lastHeartbeat = Date.now();
        }

        if ( client.invalid ) {
            client.invalid = false;
            logStatistic( client, "FP" );
        }

        do_debug_checks(client);

    } );

    socket.on( "error", function( err ) {
    	fa_server_log( 'socket.on(error,',socket.name,'): ' + err );
        logStatistic( client, "E" );
    } );

    // Remove the client from the list when it leaves
    socket.on( "end", function( ) {
    	fa_server_log( "client disconnected (" + socket.name + ")" );
        client.active = false;
        logStatistic( client, "LP" );
        logStatistic( client, "D" );
    } );

} ).listen( faServerPort );

log.info( "fa-server started listening on port " + faServerPort );

function do_debug_checks(client) {
    if (!data || 'object' != typeof data.config) {
        fa_server_log("DMC do_debug_checks() missing global.data")
        return;
    }

    // DMC: test/debugging:
    if (data.config.solo_test_huhu) {
        // var cmd = new Buffer("huhu\r\n");
        // if (typeof client.dmc_fn == 'function') {
        //     console.log("call dmc_fn");
        //     client.dmc_fn();
        // }
        // console.log("fa-server send xx huhu cmd=", cmd,"xx=",cmd.toString());
        // client.socket.write(cmd);
        fa_server_log("solo_test_huhu");
        client.fn_send_huhu();
        apply_crud({
            config: {
                solo_test_huhu: false
                ,dmc_was_here:2
            }
        });
    }
    if (data.config.solo_test_req_shots) {
        var i = data.config.solo_test_req_shots.shot_no;
        var j = data.config.solo_test_req_shots.n_shots || 1;
        client.fn_request_shots(i,j);
        // var cmd = new Buffer("send shot:") + new Buffer([i, j]);
        // console.log("Request shot#", i, "nshots=", i, "cmd=", Buffer(cmd.toString("hex")));
        // client.socket.write(cmd);
        apply_crud({
            config: {
                solo_test_req_shots: 0 // null, undefined don't work
                ,dmc_was_here:1
            }
        });
        fa_server_log("DMC test_req_shots config=",data.config.solo_test_req_shots,data.config.dmc_was_here);
    }

}

// 'client' member function
function send_huhu() {
	if (!this.socket || !this.socket.writable) {
	    log.trace("send_huhu -- socket NOT WRITABLE",this.socket.name,this.socket.writable);
		return;
	}
    var cmd = new Buffer("huhu\r\n");
    log.trace("fa-server send xx huhu cmd=", cmd, "xx=", cmd.toString());
    this.socket.write(cmd);
    
    // fetch TSOLO-C2 FW version
    
	var exec = require("child_process").exec;
	var cmd = 'ssh  -y -i /root/.ssh/id_rsa2 root@' + this.ip;
	cmd += ' fgrep C2FW /www/smtv2/server.js';
	console.log('run command: ' + cmd);
	var bind_this = this;
	exec(cmd, function(error, stdout, stderr) {
		//console.log("TSOLO C2 fwcheck exec callback");
		//console.log("err=",error);

		bind_this.c2fwversion = "";
		if (!error) {
			bind_this.c2fwversion = stdout.trim();
		}
		//console.log("this.c2fwversion =",bind_this.c2fwversion);
	});
}

// 'client' member function
function request_shots(shot_nr, n_shots) {
	if (!this.socket || !this.socket.writable) {
	    log.trace("request_shots -- socket NOT WRITABLE",this.socket.name,this.socket.writable);
		return;
	}
    var cmd = new Buffer("send shot:") + new Buffer([shot_nr, n_shots]);
    this.socket.write(cmd);
    fa_server_log("request shots #", shot_nr, "nshots=", n_shots, "cmd=", Buffer(cmd.toString("hex")));
}

// 'client' member function
function get_shot_number_mod(n) { // mod-safe
    return ((n % this.circBufSize) + this.circBufSize) % this.circBufSize;
}

// 'client' member function
function refetch_needed(incoming_shot_no) { // boolean
    // ORF sets my_client.next_shot_no
    // successfully received shot sets client.most_recent_shot_no, which will be undefined until first shot reading arrives
    //
    log.trace('check refetch_needed(',this.serialno,')...... this.most_recent_shot_no',this.most_recent_shot_no,'this.next_shot_no',this.next_shot_no);
    if (typeof this.most_recent_shot_no != 'number') return false;
    if (typeof this.next_shot_no != 'number') return false;
    if (isNaN(this.next_shot_no)) return false;
    
    if (this.circBufSize == undefined) { // then we have not received a shot since client connnection
    	// was made, and we are being called in order to process an incoming ORF.  Note that an ORF
    	// will have set this.next_shot_no
    	// Since a shot is our ONLY source of the circular buffer size, we therefore
    	// can't do sane/rational buffer-wrap computations
    	return true; // do_refetch will have special code for case this.circBufSize == undefined && !isNaN(this.next_shot_no)
    }
    
    if (typeof incoming_shot_no == 'number') { // decide if incoming shot# is the one we are waiting for?
        if (this.get_shot_number_mod(this.most_recent_shot_no+1-incoming_shot_no) == 0) {
            log.trace('refetch_needed()...... FALSE 1...incoming_shot_no is the one we were waiting for');
            return false;
        } {
            log.trace('refetch_needed()...... TRUE 1...incoming_shot_no '
            ,incoming_shot_no,"*NOT* the one we want",this.most_recent_shot_no+1);
            return true;
        }
    }

	// incoming_shot_no is not a number ==> this happens when we are called by orf_updater()
    if (this.get_shot_number_mod(this.next_shot_no-1) != this.get_shot_number_mod(this.most_recent_shot_no)) {
        log.trace("refetch_needed() TRUE 2",this.next_shot_no,this.most_recent_shot_no,
        'modcheck:',this.get_shot_number_mod(this.next_shot_no-1),this.get_shot_number_mod(this.most_recent_shot_no));
        return true; // mismatch; need to refetch
    }
    log.trace('refetch_needed()...... FALSE 2');
    return false;
}

// 'client' member function
function do_refetch() {
    if (this.circBufSize == undefined) { // then we have not received a shot since client connnection
    	// was made, and we therefore have NO idea about the size of the circular buffer, and therefore
    	// can't do sane/rational computations on it
    	// so as to discover this.circBufSize, we must ask for ANY shot at all, except for shot # = most_recent_shot_no+1
    	if (!isNaN(this.most_recent_shot_no)) {
	        this.fn_request_shots(this.most_recent_shot_no, 1);
    	} else if (undefined != this.next_shot_no ) { // if server has been restartedt hen most_recent_shot is unknown, in which case, wait until
    	} else {
    		// eventually we will get an ORF, which will set this.next_shot_no
    	}
    	return;
	}


    var n_shots_needed = this.get_shot_number_mod(
        this.next_shot_no - 1 - this.most_recent_shot_no
    );
    var max_shots_to_fetch = 15; // as many as 20
    if (n_shots_needed > max_shots_to_fetch) n_shots_needed = max_shots_to_fetch;
    var next_shot_to_get = this.get_shot_number_mod(this.most_recent_shot_no + 1);
    fa_server_log("do_refetch(), n_shots_needed =",n_shots_needed,"next_shot_to_get =", next_shot_to_get
      ,'('+this.circBufSize+')'
      , 'next_shot_no=',this.next_shot_no,'most_recent_shot_no=',this.most_recent_shot_no
    );
    if (n_shots_needed) {
        fa_server_log("do_refetch recent,next,num:", this.most_recent_shot_no, this.next_shot_no, "==>", n_shots_needed,
            "next_shot_to_get, n_shots_needed==>",next_shot_to_get, n_shots_needed);
        this.fn_request_shots(next_shot_to_get, n_shots_needed);
    }
}


setInterval( function() {
    var logStat = log4js.getLogger( "fa-server-stats.js" );
    logStat.debug( "______________________________________________________________________________" );
    logStat.debug( "Statistics:" );

	var prevClientAssoc = {}; // to calculate time lost between successive connections
	var nActive = 0;
    for ( var i = 0; i < clients.length; i++ ) {
        var c = clients[ i ];
		var prevClient = prevClientAssoc[ c.ip ];
        var msg = "";

        nActive += ( !!c.active );

        msg += ( !c.active ? "X..." : "" );
        msg += "IP: " + c.ip + ":" + c.port;
        msg += "; heartbeats: " + c.heartbeats;
        msg += "; start: " + new Date( c.start ).toLocaleTimeString();
        msg += "; active: " + Math.round( ( c.last - c.start ) / 1000 ) + "s";
        msg += "; last package: " + Math.round( ( Date.now() - c.last ) / 1000 ) + "s";
        if ( prevClient ) {
            msg += "; coverage gap: " + Math.round( ( c.start - prevClient.last ) / 1000 ) + "s";
        }
        logStat.debug( msg );
		prevClientAssoc[ c.ip ] = c;
    }
    logStat.debug( "number of created sockets: " + clients.length, "Active clients: ", nActive );

}, 10000 );

setInterval( function() {
    var now = Date.now() - 7000;
    for ( var i = 0; i < clients.length; i++ ) {
        var c = clients[ i ];
        if ( c.active && ( c.lastHeartbeat === undefined || c.lastHeartbeat < now ) && !c.invalid ) {
            logStatistic( c, "LP" );
            logStatistic( c, "C" );
            c.invalid = true;
        }
    }
}, 7000 );

var emitBroadcast = function (client) {
    var device_msg = {
        type: "SOLO",
        serialno: client.serialno, // set by ORF
        hwversion: client.hwversion, fwversion: client.fwversion, // set by huhu-ORF
        c2fwversion: client.c2fwversion,
        ping_health: client.pinger && client.pinger.get_health(),
        remote_rf_level: client.remote_rf_level,
        // os_hostname: os.hostname(), // seems unnecessary
        temperature_uncal: (typeof client.mean_temp_uncal == 'number') ?
            (+client.mean_temp_uncal.toFixed(1)) : client.mean_temp_uncal, // uncalibrated SOLO temperature_reading 
        SOLOconnected: client.connected, // RED/GREEN state? (late ORF?)
        name: "",
        shot_valid: false,
        idx: client.idx,
        deviceIP: client.ip
    };
//console.log("emitBroadcast SOLOconnected:",client.connected);
    global.qndEventEmitter.emit( "deviceBroadcastRecv", { message: device_msg, ip: client.localIp } );
};

setInterval( function() {

    for ( var i = 0; i < clients.length; i++ ) {
        var c = clients[ i ];
        if ( c.active && !c.invalid && c.device ) {
        	c.idx = i;
            emitBroadcast( c );
        }
    }

}, 3000 );

function fa_server_log() {
    if (global.data && global.data.config && global.data.config.fa_server_log) { // controlled by boolean data.config.fa_server_log
        log.debug.apply(log,Array.prototype.slice.call(arguments));
    }
	// console.log.apply(log,Array.prototype.slice.call(arguments));
}

var logStatistic = function( client, type ) {
    var logger = log4js.getLogger( "fa-server-uptimes.js" );
    logger.setLevel("INFO");
    var msg = Date.now();

    if ( type === "LP" && client.lastHeartbeat !== undefined ) {
        msg = client.lastHeartbeat;
    } else if ( type === "LP" ) {
        return;
    }

    msg += "," + client.ip;
    msg += "," + type;
    logger.info( msg );
}

function calc_mean_sd(samples) { // cloned from utils.js.... better that than another eval'ing...
	if (samples.length < 1) return {mean: 0, sd: 0, min:0, max:0, stderr:0, meanrange: 0 };
	var _avg = 0;
	var _sd = 0;
	var _mr = 0;
	var _min, _max;
	_min = _max = samples[0];
	for (i in samples) {
		_avg += samples[i];
		if (samples[i] < _min) _min = samples[i];
		if (samples[i] > _max) _max = samples[i];
	}
	_avg /= samples.length;
	for (i in samples) {
		_sd += (samples[i] - _avg)*(samples[i] - _avg);
		_mr += Math.abs(samples[i] - _avg);
	}
	_sd = Math.sqrt(_sd / samples.length);
	_mr = _mr / samples.length;
	return {mean:_avg, sd: _sd, min: _min, max: _max, stderr: _sd/Math.sqrt(samples.length),  meanrange:_mr };
}

exports.set_stuff = function(data) { // messy..
    // console.log("fa-server set_stuff()",data && data.config)
    global.data = data;
}

// l=(s=cur_string().data.shots).length, r=[];for(var i in s) r.push([parseInt(i)+1,s[i].mcucount]); JSON.stringify(r)

