/*******************************************************************************
 client_string.js: client class for a string of shots

 $Author: dmc $
 $Revision: 3206 $
 $Date: 2023-07-06 10:53:57 -0300 (Thu, 06 Jul 2023) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smtv2-lodi/public/javascripts/client_string.js $
*******************************************************************************/

// String: a series of shots on a frame, face

// everything in 'data' is synced with the server
// everything else is client-side only

function ClientString(data) {
	this.data = data;

	//log("creating string: " + data.name);
	
	// just in case server didn't send it
	if (!data.shot_params) data.shot_params = {};
	if (!data.shots) data.shots = [];

	// request update on server
	this.crud = function(delta) {
		if (undefined === this.id()) { // sanity
			return;
		}
		var msg = new_crud();
		msg.strings[this.id()] = delta;
		log("string.crud()", msg);
		g_socket.emit("crud", msg);
	}
	
	// handle crud on a specific shot
	this.crud_shot = function(i, delta) {
		var delta_shots = [];
		delta_shots[i] = delta; // TODO validshots
		this.crud({shots: delta_shots}); // TODO validshots
	}

	this.crud_shots = function(delta_shots) {
		if (is_validshots()) {
			this.crud({validshots: delta_shots});
		} else {
			this.crud({shots: delta_shots});
		}
	}

	this.id = function() {
		return this.data._id;
	}

	// assigned frame
	this.frame = function() {
		return g_frames[this.data.frame_id];
	}

	// assigned face
	this.face = function() {
		return g_faces[this.data.face_id];
	}
	
	this.validshots = function() {
		if (is_validshots()) return data.validshots;
		return data.shots;
	}
	this.shots = function() {
		return data.shots;
	}
	
	// currently selected shot (client only)
	this.cur_shot_index = -1;
	this.cur_shot = function() {
		return data.shots[this.cur_shot_index];
	}
	this.select_shot = function(index) {
		this.cur_shot_index = index;
		var shot = this.cur_shot();
		if (shot && this.shot_is_whack(shot)) { // should shotbubble go 'on'?
			// was: always yes, when starred.
			// shotbubble.enabled = true; // default shotbubble 'on' when selecting a starred shot
			// now: only auto-shown for ADMIN or RO
			if (admin_check() || ro_check()) {
				shotbubble.enabled = true; // default shotbubble 'on' when selecting a starred shot
			}
		}
	}


   this.shot_is_whack = function(shot) {// function to check if shot is possibly a problem
        if (!shot) return true;
        if (shot.is_starred) return true;
        if ((admin_check() || ro_check()) && shot.is_starred_ro) return true;
        return false;
    }
	
	// select most recent valid shot
	this.select_last_shot = function() {
		var last = this.cur_shot_index;
		for (var i = data.shots.length - 1; i >= 0; i--) {
			var shot = data.shots[i];
			if (!shot || !shot.valid) continue;
			this.cur_shot_index = i;
			break;
		}
		if (cur_string() == this && this.cur_shot_index != last) {
			shotbubble.enabled = false;
		}
		this.select_shot(this.cur_shot_index);
	}
}


